from torchquantum import tq


def random_layer_test():
    import pdb

    pdb.set_trace()
    q_layer = tq.RandomLayerAllTypes(n_ops=100, wires=[2, 1, 0], qiskit_compatible=True)
    print(q_layer)

    # q_layer = tq.RandomLayer(n_ops=100, wires=[8, 9, 10])
    # print(q_layer)

    q_layer2 = tq.RandomLayer(n_ops=200, wires=[1, 2, 3])

    q_layer2.rebuild_random_layer_from_op_list(
        n_ops_in=q_layer.n_ops, wires_in=q_layer.wires, op_list_in=q_layer.op_list
    )
    print(q_layer2)


if __name__ == "__main__":
    random_layer_test()



import torch
import torch.nn as nn
import torchquantum as tq
import torchquantum.functional as tqf
import numpy as np

from enum import IntEnum
from torchquantum.functional import mat_dict
from torchquantum.quantization.clifford_quantization import CliffordQuantizer
from abc import ABCMeta
from .macro import C_DTYPE, F_DTYPE
from torchpack.utils.logging import logger
from typing import Iterable, Union, List

__all__ = [
    "op_name_dict",
    "Operator",
    "Operation",
    "DiagonalOperation",
    "Observable",
    "Hadamard",
    "H",
    "SHadamard",
    "SH",
    "PauliX",
    "PauliY",
    "PauliZ",
    "I",
    "S",
    "T",
    "SX",
    "CNOT",
    "CZ",
    "CY",
    "RX",
    "RY",
    "RZ",
    "RXX",
    "RYY",
    "RZZ",
    "RZX",
    "SWAP",
    "SSWAP",
    "CSWAP",
    "Toffoli",
    "PhaseShift",
    "Rot",
    "MultiRZ",
    "CRX",
    "CRY",
    "CRZ",
    "CRot",
    "U1",
    "U2",
    "U3",
    "CU1",
    "CU2",
    "CU3",
    "QubitUnitary",
    "QubitUnitaryFast",
    "TrainableUnitary",
    "TrainableUnitaryStrict",
    "MultiCNOT",
    "MultiXCNOT",
    "Reset",
    "SingleExcitation",
    "EchoedCrossResonance",
    "ECR",
]


class WiresEnum(IntEnum):
    """Integer enumeration class
    to represent the number of wires
    an operation acts on."""

    AnyWires = -1
    AllWires = 0


class NParamsEnum(IntEnum):
    """Integer enumeration class
    to represent the number of wires
    an operation acts on"""

    AnyNParams = -1


AnyNParams = NParamsEnum.AnyNParams


AllWires = WiresEnum.AllWires
"""IntEnum: An enumeration which represents all wires in the
subsystem. It is equivalent to an integer with value 0."""

AnyWires = WiresEnum.AnyWires
"""IntEnum: An enumeration which represents any wires in the
subsystem. It is equivalent to an integer with value -1."""


class Operator(tq.QuantumModule):
    """The class for quantum operators."""

    fixed_ops = [
        "Hadamard",
        "SHadamard",
        "PauliX",
        "PauliY",
        "PauliZ",
        "I",
        "S",
        "T",
        "SX",
        "CNOT",
        "CZ",
        "CY",
        "SWAP",
        "SSWAP",
        "CSWAP",
        "Toffoli",
        "MultiCNOT",
        "MultiXCNOT",
        "Reset",
        "EchoedCrossResonance",
    ]

    parameterized_ops = [
        "RX",
        "RY",
        "RZ",
        "RXX",
        "RYY",
        "RZZ",
        "RZX",
        "PhaseShift",
        "Rot",
        "MultiRZ",
        "CRX",
        "CRY",
        "CRZ",
        "CRot",
        "U1",
        "U2",
        "U3",
        "CU1",
        "CU2",
        "CU3",
        "QubitUnitary",
        "QubitUnitaryFast",
        "TrainableUnitary",
        "TrainableUnitaryStrict",
        "SingleExcitation",
    ]

    @property
    def name(self):
        """String for the name of the operator."""
        return self._name

    @name.setter
    def name(self, value):
        """Set the name of the operator.

        Args:
            value (str): operator name.

        """
        self._name = value

    def __init__(
        self,
        has_params: bool = False,
        trainable: bool = False,
        init_params=None,
        n_wires=None,
        wires=None,
    ):
        """__init__ function for Operator.

        Args:
            has_params (bool, optional): Whether the operations has parameters.
                Defaults to False.
            trainable (bool, optional): Whether the parameters are trainable
                (if contains parameters). Defaults to False.
            init_params (torch.Tensor, optional): Initial parameters.
                Defaults to None.
            n_wires (int, optional): Number of qubits. Defaults to None.
            wires (Union[int, List[int]], optional): Which qubit the operation
                is applied to. Defaults to None.
        """
        super().__init__()
        self.params = None
        # number of wires of the operator
        # n_wires is used in gates that can be applied to arbitrary number
        # of qubits such as MultiRZ
        self.n_wires = n_wires
        # wires that the operator applies to
        self.wires = wires
        self._name = self.__class__.__name__
        # for static mode
        self.static_matrix = None
        self.inverse = False
        self.clifford_quantization = False

        try:
            assert not (trainable and not has_params)
        except AssertionError:
            has_params = True
            logger.warning(
                f"Module must have parameters to be trainable; "
                f"Switched 'has_params' to True."
            )

        self.has_params = has_params
        self.trainable = trainable
        if self.has_params:
            self.params = self.build_params(trainable=self.trainable)
            self.reset_params(init_params)

    @classmethod
    def _matrix(cls, params):
        """The unitary matrix of the operator.

        Args:
            params (torch.Tensor, optional): The parameters for parameterized
                operators.

        Returns: None.

        """
        raise NotImplementedError

    @property
    def matrix(self):
        """The unitary matrix of the operator."""
        return self._matrix(self.params)

    @classmethod
    def _eigvals(cls, params):
        """The eigenvalues of the unitary matrix of the operator.

        Args:
            params (torch.Tensor, optional): The parameters for parameterized
                operators.

        Returns: None.

        """
        # Warning: The eigenvalues of the operator {cls.__name__} are not defined.
        return None

    @property
    def eigvals(self):
        """The eigenvalues of the unitary matrix of the operator.

        Returns: Eigenvalues.

        """
        return self._eigvals(self.params)

    def _get_unitary_matrix(self):
        """Obtain the unitary matrix of the operator.

        Returns: Unitary matrix.

        """
        return self.matrix

    def set_wires(self, wires):
        """Set which qubits the operator is applied to.

        Args:
            wires (Union[int, List[int]]): Qubits the operator is applied to.

        Returns: None.

        """
        self.wires = [wires] if isinstance(wires, int) else wires

    def forward(
        self, q_device: tq.QuantumDevice, wires=None, params=None, inverse=False
    ):
        """Apply the operator to the quantum device states.

        Args:
            q_device (torchquantum.QuantumDevice): Quantum Device that the
                operator is applied to.
            wires (Union[int, List[int]]): Qubits that the operator is
                applied to.
            params (torch.Tensor): Parameters of the operator
            inverse (bool): Whether inverse the unitary matrix of the operator.

        Returns:

        """
        # try:
        #     assert self.name in self.fixed_ops or \
        #            self.has_params ^ (params is not None)
        # except AssertionError as err:
        #     logger.exception(f"Parameterized gate either has its "
        #                      f"own parameters or has input as parameters")
        #     raise err

        # try:
        #     assert not (self.wires is None and wires is None)
        # except AssertionError as err:
        #     logger.exception(f"Need to specify the wires either when "
        #                      f"initialize or when forward")
        #     raise err

        if params is not None:
            self.params = params

        if self.params is not None:
            self.params = (
                self.params.unsqueeze(-1) if self.params.dim() == 1 else self.params
            )

        if wires is not None:
            # update the wires
            wires = [wires] if isinstance(wires, int) else wires
            self.wires = wires

        self.inverse = inverse

        if self.static_mode:
            self.parent_graph.add_op(self)
            return

        # non-parameterized gate
        if self.params is None:
            if self.n_wires is None:
                self.func(q_device, self.wires, inverse=inverse)
            else:
                self.func(q_device, self.wires, n_wires=self.n_wires, inverse=inverse)
        else:
            if isinstance(self.noise_model_tq, tq.NoiseModelTQPhase):
                params = self.noise_model_tq.add_noise(self.params)
            else:
                params = self.params

            if self.clifford_quantization:
                params = CliffordQuantizer.quantize_sse(params)
            if self.n_wires is None:
                self.func(q_device, self.wires, params=params, inverse=inverse)
            else:
                self.func(
                    q_device,
                    self.wires,
                    params=params,
                    n_wires=self.n_wires,
                    inverse=inverse,
                )

        if self.noise_model_tq is not None and self.noise_model_tq.is_add_noise:
            noise_ops = self.noise_model_tq.sample_noise_op(self)
            if len(noise_ops):
                for noise_op in noise_ops:
                    noise_op(q_device)

    def __repr__(self):
        return f" class: {self.name} \n parameters: {self.params} \n wires: {self.wires} \n inverse: {self.inverse}"


class Observable(Operator, metaclass=ABCMeta):
    """Class for Observables."""

    def __init__(
        self,
        has_params: bool = False,
        trainable: bool = False,
        init_params=None,
        n_wires=None,
        wires=None,
    ):
        """Init function of the Observable class

        has_params (bool, optional): Whether the operations has parameters.
                Defaults to False.
            trainable (bool, optional): Whether the parameters are trainable
                (if contains parameters). Defaults to False.
            init_params (torch.Tensor, optional): Initial parameters.
                Defaults to None.
            n_wires (int, optional): Number of qubits. Defaults to None.
            wires (Union[int, List[int]], optional): Which qubit the operation
                is applied to. Defaults to None.
        """
        super().__init__(
            has_params=has_params,
            trainable=trainable,
            init_params=init_params,
            n_wires=n_wires,
            wires=wires,
        )
        self.return_type = None

    def diagonalizing_gates(self):
        """The diagonalizing gates when perform measurements.

        Returns: None.

        """
        raise NotImplementedError


class Operation(Operator, metaclass=ABCMeta):
    """_summary_"""

    def __init__(
        self,
        has_params: bool = False,
        trainable: bool = False,
        init_params=None,
        n_wires=None,
        wires=None,
    ):
        """_summary_

        Args:
            has_params (bool, optional): Whether the operations has parameters.
                Defaults to False.
            trainable (bool, optional): Whether the parameters are trainable
                (if contains parameters). Defaults to False.
            init_params (torch.Tensor, optional): Initial parameters.
                Defaults to None.
            n_wires (int, optional): Number of qubits. Defaults to None.
            wires (Union[int, List[int]], optional): Which qubit the operation is applied to.
                Defaults to None.
        """
        super().__init__(
            has_params=has_params,
            trainable=trainable,
            init_params=init_params,
            n_wires=n_wires,
            wires=wires,
        )
        if type(self.num_wires) == int:
            self.n_wires = self.num_wires

    @property
    def matrix(self):
        """The unitary matrix of the operator."""
        op_matrix = self._matrix(self.params)

        return op_matrix

    @property
    def eigvals(self):
        """ "The eigenvalues of the unitary matrix of the operator.

        Returns:
            torch.Tensor: Eigenvalues.

        """
        op_eigvals = self._eigvals(self.params)

        return op_eigvals

    def init_params(self):
        """Initialize the parameters.

        Raises:
            NotImplementedError: The init param function is not implemented.
        """
        raise NotImplementedError

    def build_params(self, trainable):
        """Build parameters.

        Args:
            trainable (bool): Whether the parameters are trainable.

        Returns:
            torch.Tensor: Built parameters.
        """
        parameters = nn.Parameter(torch.empty([1, self.num_params], dtype=F_DTYPE))
        parameters.requires_grad = True if trainable else False
        # self.register_parameter(f"{self.name}_params", parameters)
        return parameters

    def reset_params(self, init_params=None):
        """Reset parameters.

        Args:
            init_params (torch.Tensor, optional): Input the initialization
                parameters. Defaults to None.
        """
        if init_params is not None:
            if isinstance(init_params, Iterable):
                for k, init_param in enumerate(init_params):
                    torch.nn.init.constant_(self.params[:, k], init_param)
            else:
                torch.nn.init.constant_(self.params, init_params)
        else:
            torch.nn.init.uniform_(self.params, -np.pi, np.pi)


class DiagonalOperation(Operation, metaclass=ABCMeta):
    """Class for Diagonal Operation."""

    @classmethod
    def _eigvals(cls, params):
        """The eigenvalues of the unitary matrix of the operator.

        Args:
            params (torch.Tensor, optional): The parameters for parameterized
                operators.

        Returns: None.
        raise NotImplementedError
        """

    @property
    def eigvals(self):
        """The eigenvalues of the unitary matrix of the operator.

        Returns: Eigenvalues.

        """
        return super().eigvals

    @classmethod
    def _matrix(cls, params):
        """The unitary matrix of the operator.

        Args:
            params (torch.Tensor, optional): The parameters for parameterized
                operators.

        Returns: None.

        """
        return torch.diag(cls._eigvals(params))


class Hadamard(Observable, metaclass=ABCMeta):
    """Class for Hadamard Gate."""

    num_params = 0
    num_wires = 1
    eigvals = torch.tensor([1, -1], dtype=C_DTYPE)
    matrix = mat_dict["hadamard"]
    func = staticmethod(tqf.hadamard)

    @classmethod
    def _matrix(cls, params):
        return cls.matrix

    @classmethod
    def _eigvals(cls, params):
        return cls.eigvals

    def diagonalizing_gates(self):
        return [tq.RY(has_params=True, trainable=False, init_params=-np.pi / 4)]


class SHadamard(Operation, metaclass=ABCMeta):
    """Class for SHadamard Gate."""

    num_params = 0
    num_wires = 1
    matrix = mat_dict["shadamard"]
    func = staticmethod(tqf.shadamard)

    @classmethod
    def _matrix(cls, params):
        return cls.matrix


class PauliX(Observable, metaclass=ABCMeta):
    """Class for Pauli X Gate."""

    num_params = 0
    num_wires = 1
    eigvals = torch.tensor([1, -1], dtype=C_DTYPE)
    matrix = mat_dict["paulix"]
    func = staticmethod(tqf.paulix)

    @classmethod
    def _matrix(cls, params):
        return cls.matrix

    @classmethod
    def _eigvals(cls, params):
        return cls.eigvals

    def diagonalizing_gates(self):
        return [tq.Hadamard()]


class PauliY(Observable, metaclass=ABCMeta):
    """Class for Pauli Y Gate."""

    num_params = 0
    num_wires = 1
    eigvals = torch.tensor([1, -1], dtype=C_DTYPE)
    matrix = mat_dict["pauliy"]
    func = staticmethod(tqf.pauliy)

    @classmethod
    def _matrix(cls, params):
        return cls.matrix

    @classmethod
    def _eigvals(cls, params):
        return cls.eigvals

    def diagonalizing_gates(self):
        return [tq.PauliZ(), tq.S(), tq.Hadamard()]


class PauliZ(Observable, metaclass=ABCMeta):
    """Class for Pauli Z Gate."""

    num_params = 0
    num_wires = 1
    eigvals = torch.tensor([1, -1], dtype=C_DTYPE)
    matrix = mat_dict["pauliz"]
    func = staticmethod(tqf.pauliz)

    @classmethod
    def _matrix(cls, params):
        return cls.matrix

    @classmethod
    def _eigvals(cls, params):
        return cls.eigvals

    def diagonalizing_gates(self):
        return []


class I(Observable, metaclass=ABCMeta):
    """Class for Identity Gate."""

    num_params = 0
    num_wires = 1
    eigvals = torch.tensor([1, 1], dtype=C_DTYPE)
    matrix = mat_dict["i"]
    func = staticmethod(tqf.i)

    @classmethod
    def _matrix(cls, params):
        return cls.matrix

    @classmethod
    def _eigvals(cls, params):
        return cls.eigvals

    def diagonalizing_gates(self):
        return []


class S(DiagonalOperation, metaclass=ABCMeta):
    """Class for S Gate."""

    num_params = 0
    num_wires = 1
    eigvals = torch.tensor([1, 1j], dtype=C_DTYPE)
    matrix = mat_dict["s"]
    func = staticmethod(tqf.s)

    @classmethod
    def _matrix(cls, params):
        return cls.matrix

    @classmethod
    def _eigvals(cls, params):
        return cls.eigvals


class T(DiagonalOperation, metaclass=ABCMeta):
    """Class for T Gate."""

    num_params = 0
    num_wires = 1
    eigvals = torch.tensor([1, 1j], dtype=C_DTYPE)
    matrix = mat_dict["t"]
    func = staticmethod(tqf.t)

    @classmethod
    def _matrix(cls, params):
        return cls.matrix

    @classmethod
    def _eigvals(cls, params):
        return cls.eigvals


class SX(Operation, metaclass=ABCMeta):
    """Class for SX Gate."""

    num_params = 0
    num_wires = 1
    eigvals = torch.tensor([1, 1j], dtype=C_DTYPE)
    matrix = mat_dict["sx"]
    func = staticmethod(tqf.sx)

    @classmethod
    def _matrix(cls, params):
        return cls.matrix

    @classmethod
    def _eigvals(cls, params):
        return cls.eigvals


class CNOT(Operation, metaclass=ABCMeta):
    """Class for CNOT Gate."""

    num_params = 0
    num_wires = 2
    matrix = mat_dict["cnot"]
    func = staticmethod(tqf.cnot)

    @classmethod
    def _matrix(cls, params):
        return cls.matrix


class CZ(DiagonalOperation, metaclass=ABCMeta):
    """Class for CZ Gate."""

    num_params = 0
    num_wires = 2
    eigvals = np.array([1, 1, 1, -1])
    matrix = mat_dict["cz"]
    func = staticmethod(tqf.cz)

    @classmethod
    def _matrix(cls, params):
        return cls.matrix

    @classmethod
    def _eigvals(cls, params):
        return cls.eigvals


class CY(Operation, metaclass=ABCMeta):
    """Class for CY Gate."""

    num_params = 0
    num_wires = 2
    matrix = mat_dict["cy"]
    func = staticmethod(tqf.cy)

    @classmethod
    def _matrix(cls, params):
        return cls.matrix


class SWAP(Operation, metaclass=ABCMeta):
    """Class for SWAP Gate."""

    num_params = 0
    num_wires = 2
    matrix = mat_dict["swap"]
    func = staticmethod(tqf.swap)

    @classmethod
    def _matrix(cls, params):
        return cls.matrix


class SSWAP(Operation, metaclass=ABCMeta):
    """Class for SSWAP Gate."""

    num_params = 0
    num_wires = 2
    matrix = mat_dict["sswap"]
    func = staticmethod(tqf.sswap)

    @classmethod
    def _matrix(cls, params):
        return cls.matrix


class CSWAP(Operation, metaclass=ABCMeta):
    """Class for CSWAP Gate."""

    num_params = 0
    num_wires = 3
    matrix = mat_dict["cswap"]
    func = staticmethod(tqf.cswap)

    @classmethod
    def _matrix(cls, params):
        return cls.matrix


class Toffoli(Operation, metaclass=ABCMeta):
    """Class for Toffoli Gate."""

    num_params = 0
    num_wires = 3
    matrix = mat_dict["toffoli"]
    func = staticmethod(tqf.toffoli)

    @classmethod
    def _matrix(cls, params):
        return cls.matrix


class RX(Operation, metaclass=ABCMeta):
    """Class for RX Gate."""

    num_params = 1
    num_wires = 1
    func = staticmethod(tqf.rx)

    @classmethod
    def _matrix(cls, params):
        return tqf.rx_matrix(params)


class RY(Operation, metaclass=ABCMeta):
    """Class for RY Gate."""

    num_params = 1
    num_wires = 1
    func = staticmethod(tqf.ry)

    @classmethod
    def _matrix(cls, params):
        return tqf.ry_matrix(params)


class RZ(DiagonalOperation, metaclass=ABCMeta):
    """Class for RZ Gate."""

    num_params = 1
    num_wires = 1
    func = staticmethod(tqf.rz)

    @classmethod
    def _matrix(cls, params):
        return tqf.rz_matrix(params)


class PhaseShift(DiagonalOperation, metaclass=ABCMeta):
    """Class for PhaseShift Gate."""

    num_params = 1
    num_wires = 1
    func = staticmethod(tqf.phaseshift)

    @classmethod
    def _matrix(cls, params):
        return tqf.phaseshift_matrix(params)


class Rot(Operation, metaclass=ABCMeta):
    """Class for Rotation Gate."""

    num_params = 3
    num_wires = 1
    func = staticmethod(tqf.rot)

    @classmethod
    def _matrix(cls, params):
        return tqf.rot_matrix(params)


class MultiRZ(DiagonalOperation, metaclass=ABCMeta):
    """Class for Multi-qubit RZ Gate."""

    num_params = 1
    num_wires = AnyWires
    func = staticmethod(tqf.multirz)

    @classmethod
    def _matrix(cls, params, n_wires):
        return tqf.multirz_matrix(params, n_wires)


class RXX(Operation, metaclass=ABCMeta):
    """Class for RXX Gate."""

    num_params = 1
    num_wires = 2
    func = staticmethod(tqf.rxx)

    @classmethod
    def _matrix(cls, params):
        return tqf.rxx_matrix(params)


class RYY(Operation, metaclass=ABCMeta):
    """Class for RYY Gate."""

    num_params = 1
    num_wires = 2
    func = staticmethod(tqf.ryy)

    @classmethod
    def _matrix(cls, params):
        return tqf.ryy_matrix(params)


class RZZ(DiagonalOperation, metaclass=ABCMeta):
    """Class for RZZ Gate."""

    num_params = 1
    num_wires = 2
    func = staticmethod(tqf.rzz)

    @classmethod
    def _matrix(cls, params):
        return tqf.rzz_matrix(params)


class RZX(Operation, metaclass=ABCMeta):
    """Class for RZX Gate."""

    num_params = 1
    num_wires = 2
    func = staticmethod(tqf.rzx)

    @classmethod
    def _matrix(cls, params):
        return tqf.rzx_matrix(params)


class TrainableUnitary(Operation, metaclass=ABCMeta):
    """Class for TrainableUnitary Gate."""

    num_params = AnyNParams
    num_wires = AnyWires
    func = staticmethod(tqf.qubitunitaryfast)

    def build_params(self, trainable):
        """Build the parameters for the gate.

        Args:
            trainable (bool): Whether the parameters are trainble.

        Returns:
            torch.Tensor: Parameters.

        """
        parameters = nn.Parameter(
            torch.empty(1, 2**self.n_wires, 2**self.n_wires, dtype=C_DTYPE)
        )
        parameters.requires_grad = True if trainable else False
        # self.register_parameter(f"{self.name}_params", parameters)
        return parameters

    def reset_params(self, init_params=None):
        """Reset the parameters.

        Args:
            init_params (torch.Tensor, optional): Initial parameters.

        Returns:
            None.

        """
        mat = torch.randn((1, 2**self.n_wires, 2**self.n_wires), dtype=C_DTYPE)
        U, Sigma, V = torch.svd(mat)
        self.params.data.copy_(U.matmul(V.permute(0, 2, 1)))

    @staticmethod
    def _matrix(self, params):
        return tqf.qubitunitaryfast(params)


class TrainableUnitaryStrict(TrainableUnitary, metaclass=ABCMeta):
    """Class for Strict Unitary matrix gate."""

    num_params = AnyNParams
    num_wires = AnyWires
    func = staticmethod(tqf.qubitunitarystrict)


class CRX(Operation, metaclass=ABCMeta):
    """Class for Controlled Rotation X gate."""

    num_params = 1
    num_wires = 2
    func = staticmethod(tqf.crx)

    @classmethod
    def _matrix(cls, params):
        return tqf.crx_matrix(params)


class CRY(Operation, metaclass=ABCMeta):
    """Class for Controlled Rotation Y gate."""

    num_params = 1
    num_wires = 2
    func = staticmethod(tqf.cry)

    @classmethod
    def _matrix(cls, params):
        return tqf.cry_matrix(params)


class CRZ(Operation, metaclass=ABCMeta):
    """Class for Controlled Rotation Z gate."""

    num_params = 1
    num_wires = 2
    func = staticmethod(tqf.crz)

    @classmethod
    def _matrix(cls, params):
        return tqf.crz_matrix(params)


class CRot(Operation, metaclass=ABCMeta):
    """Class for Controlled Rotation gate."""

    num_params = 3
    num_wires = 2
    func = staticmethod(tqf.crot)

    @classmethod
    def _matrix(cls, params):
        return tqf.crot_matrix(params)


class U1(DiagonalOperation, metaclass=ABCMeta):
    """Class for Controlled Rotation Y gate.  U1 is the same
    as phaseshift.
    """

    num_params = 1
    num_wires = 1
    func = staticmethod(tqf.u1)

    @classmethod
    def _matrix(cls, params):
        return tqf.u1_matrix(params)


class CU1(DiagonalOperation, metaclass=ABCMeta):
    """Class for controlled U1 gate."""

    num_params = 1
    num_wires = 2
    func = staticmethod(tqf.cu1)

    @classmethod
    def _matrix(cls, params):
        return tqf.cu1_matrix(params)


class U2(Operation, metaclass=ABCMeta):
    """Class for U2 gate."""

    num_params = 2
    num_wires = 1
    func = staticmethod(tqf.u2)

    @classmethod
    def _matrix(cls, params):
        return tqf.u2_matrix(params)


class CU2(Operation, metaclass=ABCMeta):
    """Class for controlled U2 gate."""

    num_params = 2
    num_wires = 2
    func = staticmethod(tqf.cu2)

    @classmethod
    def _matrix(cls, params):
        return tqf.cu2_matrix(params)


class U3(Operation, metaclass=ABCMeta):
    """Class for U3 gate."""

    num_params = 3
    num_wires = 1
    func = staticmethod(tqf.u3)

    @classmethod
    def _matrix(cls, params):
        return tqf.u3_matrix(params)


class CU3(Operation, metaclass=ABCMeta):
    """Class for Controlled U3 gate."""

    num_params = 3
    num_wires = 2
    func = staticmethod(tqf.cu3)

    @classmethod
    def _matrix(cls, params):
        return tqf.cu3_matrix(params)


class QubitUnitary(Operation, metaclass=ABCMeta):
    """Class for controlled Qubit Unitary gate."""

    num_params = AnyNParams
    num_wires = AnyWires
    func = staticmethod(tqf.qubitunitary)

    @classmethod
    def _matrix(cls, params):
        return tqf.qubitunitary_matrix(params)

    def build_params(self, trainable):
        return None

    def reset_params(self, init_params=None):
        self.params = torch.tensor(init_params, dtype=C_DTYPE)
        self.register_buffer(f"{self.name}_unitary", self.params)


class QubitUnitaryFast(Operation, metaclass=ABCMeta):
    """Class for fast implementation of
    controlled Qubit Unitary gate."""

    num_params = AnyNParams
    num_wires = AnyWires
    func = staticmethod(tqf.qubitunitaryfast)

    def __init__(
        self,
        has_params: bool = False,
        trainable: bool = False,
        init_params=None,
        n_wires=None,
        wires=None,
    ):
        super().__init__(
            has_params=True,
            trainable=trainable,
            init_params=init_params,
            n_wires=n_wires,
            wires=wires,
        )

    @classmethod
    def from_controlled_operation(
        cls,
        op,
        c_wires,
        t_wires,
        trainable,
    ):
        """

        Args:
            op: the operation
            c_wires: controlled wires, will only be a set such as 1, [2,3]
            t_wires: can be a list of list of wires, multiple sets
            [[1,2], [3,4]]
            trainable:
        """
        op = op
        c_wires = np.array(c_wires)
        t_wires = np.array(t_wires)
        trainable = trainable
        # self.n_t_wires = op.n_wires
        # assert len(t_wires) == op.n_wires

        orig_u = op.matrix
        orig_u_n_wires = op.n_wires

        wires = []

        if c_wires.ndim == 0:
            # only one control qubit
            # 1
            n_c_wires = 1
            wires.append(c_wires.item())
        elif c_wires.ndim == 1:
            # multiple control qubits
            # [1, 2]
            n_c_wires = c_wires.shape[0]
            wires.extend(list(c_wires))

        if t_wires.ndim == 0:
            # single qubit U on one set
            # 2
            n_t_wires = 1
            n_set_t_wires = 1
            wires.append(t_wires.item())
        elif t_wires.ndim == 1:
            # single qubit U on multiple sets
            # [1, 2, 3]
            # or multi qubit U on one set
            # [2, 3]
            n_t_wires = t_wires.shape[0]
            n_set_t_wires = n_t_wires // orig_u_n_wires
            wires.extend(list(t_wires.flatten()))

        elif t_wires.ndim == 2:
            # multi qubit unitary on multiple sets
            # [[2, 3], [4, 5]]
            n_t_wires = t_wires.flatten().shape[0]
            n_set_t_wires = n_t_wires // orig_u_n_wires
            wires.extend(list(t_wires.flatten()))

        n_wires = n_c_wires + n_t_wires

        # compute the new unitary, then permute
        unitary = torch.tensor(torch.zeros(2**n_wires, 2**n_wires, dtype=C_DTYPE))
        for k in range(2**n_wires - 2**n_t_wires):
            unitary[k, k] = 1.0 + 0.0j

        # compute kronecker product of all the controlled target

        controlled_u = None
        for k in range(n_set_t_wires):
            if controlled_u is None:
                controlled_u = orig_u
            else:
                controlled_u = torch.kron(controlled_u, orig_u)

        d_controlled_u = controlled_u.shape[-1]
        unitary[-d_controlled_u:, -d_controlled_u:] = controlled_u

        return cls(
            has_params=True,
            trainable=trainable,
            init_params=unitary,
            n_wires=n_wires,
            wires=wires,
        )

    @classmethod
    def _matrix(cls, params):
        return tqf.qubitunitaryfast_matrix(params)

    def build_params(self, trainable):
        return None

    def reset_params(self, init_params=None):
        self.params = torch.tensor(init_params, dtype=C_DTYPE)
        self.register_buffer(f"{self.name}_unitary", self.params)


class MultiCNOT(Operation, metaclass=ABCMeta):
    """Class for Multi qubit CNOT gate."""

    num_params = 0
    num_wires = AnyWires
    func = staticmethod(tqf.multicnot)

    @classmethod
    def _matrix(cls, params, n_wires):
        return tqf.multicnot_matrix(n_wires)

    @property
    def matrix(self):
        op_matrix = self._matrix(self.params, self.n_wires)
        return op_matrix


class MultiXCNOT(Operation, metaclass=ABCMeta):
    """Class for Multi qubit XCNOT gate."""

    num_params = 0
    num_wires = AnyWires
    func = staticmethod(tqf.multixcnot)

    @classmethod
    def _matrix(cls, params, n_wires):
        return tqf.multixcnot_matrix(n_wires)

    @property
    def matrix(self):
        op_matrix = self._matrix(self.params, self.n_wires)
        return op_matrix


class Reset(Operator, metaclass=ABCMeta):
    """Class for Reset gate."""

    num_params = 0
    num_wires = AnyWires
    func = staticmethod(tqf.reset)

    @classmethod
    def _matrix(cls, params):
        return None


class SingleExcitation(Operator, metaclass=ABCMeta):
    """Class for SingleExcitation gate."""

    num_params = 1
    num_wires = 2
    func = staticmethod(tqf.singleexcitation)

    @classmethod
    def _matrix(cls, params):
        return tqf.singleexcitation_matrix(params)


class ECR(Operation, metaclass=ABCMeta):
    """Class for Echoed Cross Resonance Gate."""

    num_params = 0
    num_wires = 2
    matrix = mat_dict["ecr"]
    func = staticmethod(tqf.ecr)

    @classmethod
    def _matrix(cls, params):
        return cls.matrix


H = Hadamard
SH = SHadamard
EchoedCrossResonance = ECR


op_name_dict = {
    "hadamard": Hadamard,
    "h": Hadamard,
    "shadamard": SHadamard,
    "sh": SHadamard,
    "paulix": PauliX,
    "x": PauliX,
    "pauliy": PauliY,
    "y": PauliY,
    "pauliz": PauliZ,
    "z": PauliZ,
    "i": I,
    "s": S,
    "t": T,
    "sx": SX,
    "cx": CNOT,
    "cnot": CNOT,
    "cz": CZ,
    "cy": CY,
    "rx": RX,
    "ry": RY,
    "rz": RZ,
    "rxx": RXX,
    "xx": RXX,
    "ryy": RYY,
    "yy": RYY,
    "rzz": RZZ,
    "zz": RZZ,
    "rzx": RZX,
    "zx": RZX,
    "swap": SWAP,
    "sswap": SSWAP,
    "cswap": CSWAP,
    "toffoli": Toffoli,
    "ccx": Toffoli,
    "phaseshift": PhaseShift,
    "rot": Rot,
    "multirz": MultiRZ,
    "crx": CRX,
    "cry": CRY,
    "crz": CRZ,
    "crot": CRot,
    "u1": U1,
    "p": U1,
    "u2": U2,
    "u3": U3,
    "u": U3,
    "cu1": CU1,
    "cp": CU1,
    "cr": CU1,
    "cphase": CU1,
    "cu2": CU2,
    "cu3": CU3,
    "cu": CU3,
    "qubitunitary": QubitUnitary,
    "qubitunitarystrict": QubitUnitaryFast,
    "qubitunitaryfast": QubitUnitaryFast,
    "trainableunitary": TrainableUnitary,
    "trainableunitarystrict": TrainableUnitaryStrict,
    "multicnot": MultiCNOT,
    "multixcnot": MultiXCNOT,
    "reset": Reset,
    "singleexcitation": SingleExcitation,
    "ecr": ECR,
    "echoedcrossresonance": ECR,
}



import torch
import torch.nn as nn

from torchpack.utils.config import configs
from torchpack.utils.typing import Dataset, Optimizer, Scheduler
from torchpack.callbacks import (InferenceRunner, MeanAbsoluteError,
                                 MaxSaver, MinSaver,
                                 Saver, SaverRestore, CategoricalAccuracy)
from examples.gradient_pruning.callbacks import LegalInferenceRunner, SubnetInferenceRunner, \
    NLLError, TrainerRestore, AddNoiseInferenceRunner, GradRestore
from torchquantum.plugins import QiskitProcessor
from torchquantum.vqe_utils import parse_hamiltonian_file
from torchquantum.noise_model import *

__all__ = ['make_optimizer',
    'make_scheduler', 'make_callbacks', 'make_qiskit_processor',
    'make_noise_model_tq'
]

def make_optimizer(model: nn.Module) -> Optimizer:
    if configs.optimizer.name == 'sgd':
        optimizer = torch.optim.SGD(
            model.parameters(),
            lr=configs.optimizer.lr,
            momentum=configs.optimizer.momentum,
            weight_decay=configs.optimizer.weight_decay,
            nesterov=configs.optimizer.nesterov)
    elif configs.optimizer.name == 'adam':
        optimizer = torch.optim.Adam(
            model.parameters(),
            lr=configs.optimizer.lr,
            weight_decay=configs.optimizer.weight_decay)
    elif configs.optimizer.name == 'adam_group':
        optimizer = torch.optim.Adam(
            [param for name, param in model.named_parameters() if
                'lambda' not in name],
            lr=configs.optimizer.lr,
            weight_decay=configs.optimizer.weight_decay)
        optimizer.add_param_group({
            'params': [param for name, param in model.named_parameters() if
                       'lambda' in name],
            'lr': -configs.optimizer.lambda_lr,
        })
    elif configs.optimizer.name == 'adamw':
        optimizer = torch.optim.AdamW(
            model.parameters(),
            lr=configs.optimizer.lr,
            weight_decay=configs.optimizer.weight_decay)
    else:
        raise NotImplementedError(configs.optimizer.name)

    return optimizer


def make_scheduler(optimizer: Optimizer) -> Scheduler:
    if configs.scheduler.name == 'constant':
        scheduler = torch.optim.lr_scheduler.LambdaLR(
            optimizer, lr_lambda=lambda epoch: 1)
    elif configs.scheduler.name == 'cosine':
        scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(
            optimizer, T_max=configs.run.n_epochs)
    elif configs.scheduler.name == 'cosine_warm':
        from .schedulers import CosineAnnealingWarmupRestarts
        scheduler = CosineAnnealingWarmupRestarts(
            optimizer,
            first_cycle_steps=configs.run.n_epochs,
            max_lr=configs.optimizer.lr,
            min_lr=configs.optimizer.min_lr,
            warmup_steps=configs.run.n_warm_epochs,
        )
    else:
        raise NotImplementedError(configs.scheduler.name)

    return scheduler


def make_trainer(model: nn.Module,
                 criterion: nn.Module,
                 optimizer: Optimizer,
                 scheduler: Scheduler
                 ):
    if configs.trainer.name == 'params_shift_trainer':
        from examples.gradient_pruning.trainers import ParamsShiftTrainer
        trainer = ParamsShiftTrainer(model=model,
                           criterion=criterion,
                           optimizer=optimizer,
                           scheduler=scheduler)
    elif configs.trainer.name == 'q_trainer':
        from examples.gradient_pruning.trainers import QTrainer
        trainer = QTrainer(model=model,
                           criterion=criterion,
                           optimizer=optimizer,
                           scheduler=scheduler)
    else:
        raise NotImplementedError(configs.trainer.name)

    return trainer


def get_subcallbacks(config):
    subcallbacks = []
    for subcallback in config:
        if subcallback['metrics'] == 'CategoricalAccuracy':
            subcallbacks.append(
                CategoricalAccuracy(name=subcallback['name'])
            )
        elif subcallback['metrics'] == 'MeanAbsoluteError':
            subcallbacks.append(
                MeanAbsoluteError(name=subcallback['name'])
            )
        elif subcallback['metrics'] == 'NLLError':
            subcallbacks.append(
                NLLError(name=subcallback['name'])
            )
        elif subcallback['metrics'] == 'MinError':
            from examples.gradient_pruning.callbacks import MinError
            subcallbacks.append(
                MinError(name=subcallback['name'])
            )
        else:
            raise NotImplementedError(subcallback['metrics'])
    return subcallbacks


def make_callbacks(dataflow, state=None):
    callbacks = []
    for config in configs['callbacks']:
        if config['callback'] == 'InferenceRunner':
            callback = InferenceRunner(
                dataflow=dataflow[config['split']],
                callbacks=get_subcallbacks(config['subcallbacks'])
            )
        elif config['callback'] == 'LegalInferenceRunner':
            callback = LegalInferenceRunner(
                dataflow=dataflow[config['split']],
                callbacks=get_subcallbacks(config['subcallbacks'])
            )
        elif config['callback'] == 'SubnetInferenceRunner':
            callback = SubnetInferenceRunner(
                dataflow=dataflow[config['split']],
                callbacks=get_subcallbacks(config['subcallbacks']),
                subnet=config['subnet']
            )
        elif config['callback'] == 'AddNoiseInferenceRunner':
            callback = AddNoiseInferenceRunner(
                dataflow=dataflow[config['split']],
                callbacks=get_subcallbacks(config['subcallbacks']),
                noise_total_prob=config['noise_total_prob']
            )
        elif config['callback'] == 'SaverRestore':
            callback = SaverRestore()
        elif config['callback'] == 'Saver':
            callback = Saver(max_to_keep=config['max_to_keep'])
        elif config['callback'] == 'MaxSaver':
            callback = MaxSaver(config['name'])
        elif config['callback'] == 'MinSaver':
            callback = MinSaver(config['name'])
        elif config['callback'] == 'GradRestore':
            callback = GradRestore()
        else:
            raise NotImplementedError(config['callback'])
        callbacks.append(callback)

    if configs.ckpt.load_trainer:
        assert state is not None
        callback = TrainerRestore(state)
        callbacks.append(callback)

    return callbacks


def make_qiskit_processor():
    processor = QiskitProcessor(
        use_real_qc=configs.qiskit.use_real_qc,
        backend_name=configs.qiskit.backend_name,
        noise_model_name=configs.qiskit.noise_model_name,
        coupling_map_name=getattr(configs.qiskit, 'coupling_map_name',
                                  configs.qiskit.noise_model_name),
        basis_gates_name=configs.qiskit.noise_model_name,
        n_shots=configs.qiskit.n_shots,
        initial_layout=configs.qiskit.initial_layout,
        seed_transpiler=configs.qiskit.seed_transpiler,
        seed_simulator=configs.qiskit.seed_simulator,
        optimization_level=configs.qiskit.optimization_level,
        max_jobs=configs.qiskit.max_jobs,
        remove_ops=configs.prune.eval.remove_ops,
        remove_ops_thres=configs.prune.eval.remove_ops_thres,
        transpile_with_ancilla=getattr(configs.qiskit, 'transpile_with_ancilla', True),
        hub=getattr(configs.qiskit, 'hub', None),
        layout_method=getattr(configs.qiskit, 'layout_method', None),
        routing_method=getattr(configs.qiskit, 'routing_method', None),
    )
    return processor


def make_noise_model_tq():
    if configs.trainer.noise_model_tq_name == 'from_qiskit_read':
        noise_model_tq = NoiseModelTQ(
            noise_model_name=configs.qiskit.noise_model_name,
            n_epochs=configs.run.n_epochs,
            noise_total_prob=getattr(configs.trainer, 'noise_total_prob',
                                     None),
            ignored_ops=configs.trainer.ignored_noise_ops,
            prob_schedule=getattr(configs.trainer, 'noise_prob_schedule',
                                  None),
            prob_schedule_separator=getattr(
                configs.trainer, 'noise_prob_schedule_separator', None),
            factor=getattr(configs.trainer, 'noise_factor', None),
            add_thermal=getattr(configs.trainer, 'noise_add_thermal', True)
        )
    elif configs.trainer.noise_model_tq_name == 'from_qiskit':
        noise_model_tq = NoiseModelTQQErrorOnly(
            noise_model_name=configs.qiskit.noise_model_name,
            n_epochs=configs.run.n_epochs,
            noise_total_prob=getattr(configs.trainer, 'noise_total_prob',
                                     None),
            ignored_ops=configs.trainer.ignored_noise_ops,
            prob_schedule=getattr(configs.trainer, 'noise_prob_schedule',
                                  None),
            prob_schedule_separator=getattr(
                configs.trainer, 'noise_prob_schedule_separator', None),
            factor=getattr(configs.trainer, 'noise_factor', None),
            add_thermal=getattr(configs.trainer, 'noise_add_thermal', True)
        )
    elif configs.trainer.noise_model_tq_name == 'activation':
        noise_model_tq = NoiseModelTQActivation(
            mean=configs.trainer.noise_mean,
            std=configs.trainer.noise_std,
            n_epochs=configs.run.n_epochs,
            prob_schedule=getattr(configs.trainer, 'noise_prob_schedule',
                                  None),
            prob_schedule_separator=getattr(
                configs.trainer, 'noise_prob_schedule_separator', None),
            after_norm=getattr(configs.trainer, 'noise_after_norm',
                               False),
            factor=getattr(configs.trainer, 'noise_factor', None)
        )
    elif configs.trainer.noise_model_tq_name == 'phase':
        noise_model_tq = NoiseModelTQPhase(
            mean=configs.trainer.noise_mean,
            std=configs.trainer.noise_std,
            n_epochs=configs.run.n_epochs,
            prob_schedule=getattr(configs.trainer, 'noise_prob_schedule',
                                  None),
            prob_schedule_separator=getattr(
                configs.trainer, 'noise_prob_schedule_separator', None),
            factor=getattr(configs.trainer, 'noise_factor', None)
        )
    elif configs.trainer.noise_model_tq_name == 'only_read':
        noise_model_tq = NoiseModelTQReadoutOnly(
            noise_model_name=configs.qiskit.noise_model_name,
            n_epochs=configs.run.n_epochs,
            prob_schedule=getattr(configs.trainer, 'noise_prob_schedule',
                                  None),
            prob_schedule_separator=getattr(
                configs.trainer, 'noise_prob_schedule_separator', None),
            factor=getattr(configs.trainer, 'noise_factor', None)
        )
    elif configs.trainer.noise_model_tq_name == 'activation_read':
        noise_model_tq = NoiseModelTQActivationReadout(
            noise_model_name=configs.qiskit.noise_model_name,
            mean=configs.trainer.noise_mean,
            std=configs.trainer.noise_std,
            n_epochs=configs.run.n_epochs,
            prob_schedule=getattr(configs.trainer, 'noise_prob_schedule',
                                  None),
            prob_schedule_separator=getattr(
                configs.trainer, 'noise_prob_schedule_separator', None),
            after_norm=getattr(configs.trainer, 'noise_after_norm',
                               False),
            factor=getattr(configs.trainer, 'noise_factor', None)
        )
    elif configs.trainer.noise_model_tq_name == 'phase_read':
        noise_model_tq = NoiseModelTQPhaseReadout(
            noise_model_name=configs.qiskit.noise_model_name,
            mean=configs.trainer.noise_mean,
            std=configs.trainer.noise_std,
            n_epochs=configs.run.n_epochs,
            prob_schedule=getattr(configs.trainer, 'noise_prob_schedule',
                                  None),
            prob_schedule_separator=getattr(
                configs.trainer, 'noise_prob_schedule_separator', None),
            factor=getattr(configs.trainer, 'noise_factor', None)
        )
    else:
        raise NotImplementedError(configs.trainer.noise_model_tq_name)

    return noise_model_tq



import imp
import torch
import torch.nn as nn
import numpy as np
import functools
import torchquantum.functional as tqf
import torchquantum.Dfunc as dfunc
import torchquantum as tq
import copy
from torchquantum.states import QuantumState
from torchquantum.macro import C_DTYPE, ABC, ABC_ARRAY, INV_SQRT2
from typing import Union, List, Iterable


__all__ = ["DensityMatrix"]


class DensityMatrix(nn.Module):
    def __init__(self, n_wires: int, bsz: int = 1):
        """Init function for DensityMatrix class(Density Operator)
        Args:
            n_wires (int): how many qubits for the densityMatrix.
        """
        super().__init__()

        self.n_wires = n_wires
        """
        For example, when n_wires=3
        matrix[001110] denotes the index of |001><110|=|index1><index2|
        Set Initial value the density matrix of the pure state |00...00>
        """
        _matrix = torch.zeros(2 ** (2 * self.n_wires), dtype=C_DTYPE)
        _matrix[0] = 1 + 0j
        _matrix = torch.reshape(_matrix, [2] * (2 * self.n_wires))
        self.register_buffer("matrix", _matrix)

        repeat_times = [bsz] + [1] * len(self.matrix.shape)
        self._matrix = self.matrix.repeat(*repeat_times)
        self.register_buffer("matrix", self._matrix)

        """
        Whether or not calculate by states
        """
        self._calc_by_states = True

        """
        Remember whether or not a standard matrix on a given wire is contructed
        """
        self.construct = {}
        for key in tqf.func_name_dict.keys():
            self.construct[key] = [False] * n_wires

        """
        Store the constructed operator matrix
        """
        self.operator_matrix = {}
        for key in tqf.func_name_dict.keys():
            self.operator_matrix[key] = {}

        """
        Preserve the probability of all pure states. has the form [(p1,s1),(p2,s2),(p3,s3),...]
              2**n 2**n 2**n
         Matrix  3 purestate
        """
        self.state_list = []
        for i in range(0, bsz):
            self.state_list.append((1, QuantumState(n_wires)))

    def set_calc_by_states(self, val):
        self._calc_by_states = val

    def update_matrix_from_states(self):
        """Update the density matrix value from all pure states"""
        _matrix = torch.zeros(2 ** (2 * self.n_wires), dtype=C_DTYPE)
        _matrix = torch.reshape(_matrix, [2**self.n_wires, 2**self.n_wires])
        self.register_buffer("matrix", _matrix)
        bsz = self.matrix.shape[0]
        repeat_times = [bsz] + [1] * len(self.matrix.shape)
        self._matrix = self.matrix.repeat(*repeat_times)
        for i in range(0, bsz):
            for p, state in self.state_list:
                self._matrix[i] = self._matrix[i] + p * state.density_matrix()[0][:][:]
        self.register_buffer("matrix", self._matrix)

    def vector(self):
        return torch.reshape(_matrix, [2 ** (2 * self.n_wires)])

    def print_2d(self, index):
        """Print the matrix value of matrix[index]"""
        _matrix = torch.reshape(self._matrix[index], [2**self.n_wires] * 2)
        print(_matrix)

    def trace(self, index):
        """Return the trace of the DensityMatrix of matrix[index]"""
        return torch.trace(self._matrix[index])

    def positive_semidefinite(self, index):
        """Check whether the matrix is positive semidefinite by Sylvester's_criterion"""
        return np.all(np.linalg.eigvals(self._matrix[index]) > 0)

    def check_valid(self):
        """Check whether the matrix has trace 1 and is positive semidefinite"""
        for i in range(0, self._matrix.shape[0]):
            if self.trace(i) != 1 or not self.positive_semidefinite(i):
                return False
        return True

    def spectral(self, index):
        """Return the spectral of the DensityMatrix"""
        return list(np.linalg.eigvals(self._matrix[index]))

    def tensor(self, other):
        """Return self tensor other(Notice the order)
        Args:
            other (DensityMatrix: Another density matrix
        """
        self._matrix = torch.kron(self._matrix, other._matrix)

    def expand(self, other):
        """Return other tensor self(Notice the order)
        Args:
            other (DensityMatrix: Another density matrix
        """
        self._matrix = torch.kron(other._matrix, self._matrix)

    def clone_matrix(self, existing_matrix: torch.Tensor):
        self._matrix = existing_matrix.clone()

    def set_matrix(self, matrix: Union[torch.tensor, List]):
        matrix = torch.tensor(matrix, dtype=C_DTYPE).to(self.matrix.device)
        bsz = matrix.shape[0]
        self.matrix = torch.reshape(
            matrix, [bsz] + [2 ** (2 * self.n_wires), 2 ** (2 * self.n_wires)]
        )

    def evolve(self, operator):
        """Evolve the density matrix in batchmode
        operator has size [2**(2*self.n_wires),2**(2*self.n_wires)]
        """
        """Convert the matrix to vector of shape [bsz,2**(2*self.n_wires)]
           Return U rho U^\dagger
        """
        bsz = self.matrix.shape[0]
        expand_shape = [bsz] + list(operator.shape)

        new_matrix = operator.expand(expand_shape).bmm(
            torch.reshape(self.matrix, [bsz, 2 ** (2 * self.n_wires)])
        )
        self.matrix = torch.reshape(
            new_matrix, [bsz, 2**self.n_wires, 2**self.n_wires]
        )

    def expectation(self):
        """Expectation of a measurement"""
        return

    def set_from_state(self, probs, states: Union[torch.Tensor, List]):
        """Get the density matrix of a mixed state.
        Args:
          probs:List of probability of each state
          states:List of state.
        For example:
          probs:[0.5,0.5],states:[|00>,|11>]
        Then the corresponding matrix is: 0.5|00><00|+0.5|11><11|
         0.5, 0, 0, 0
         0  , 0, 0, 0
         0  , 0, 0, 0
         0 ,  0, 0, 0.5
         self._matrix[00][00]=self._matrix[11][11]=0.5
        """
        for i in range(0, len(probs)):
            self.state_list
        _matrix = torch.zeros(2 ** (2 * self.n_wires), dtype=C_DTYPE)
        _matrix = torch.reshape(_matrix, [2**self.n_wires, 2**self.n_wires])
        for i in range(0, len(probs)):
            row = torch.reshape(states[i], [2**self.n_wires, 1])
            col = torch.reshape(states[i], [1, 2**self.n_wires])
            _matrix = _matrix + probs[i] * torch.matmul(row, col)
        self.matrix = torch.reshape(_matrix, [2] * (2 * self.n_wires))
        return

    def _add(self, other):
        """Return self + other
        Args:
            other (complex): a complex number.
        """
        if not isinstance(other, DensityMatrix):
            other = DensityMatrix(other)
        if not self._matrix.shape == other._matrix.shape:
            raise ("Two density matrix must have the same shape.")
        ret = copy.copy(self)
        ret._matrix = self.matrix + other._matrix
        return ret

    def _multiply(self, other):
        """Return other * self.
        Args:
            other (complex): a complex number.
        """
        ret = copy.copy(self)
        ret._matrix = other * self._matrix
        return ret

    def purity(self):
        """Calculate the purity of the DensityMatrix defined as \gamma=tr(\rho^2)"""
        return torch.trace(torch.matmul(self._matrix, self._matrix))

    def partial_trace(self, dims: List[int]):
        """Calculate the partial trace of given sub-dimension, return a new density_matrix
        Args:
            dims:The list of sub-dimension
            For example, If we have 3 qubit, the matrix shape is (8,8),
            We want to do partial trace to qubit 0,2, dims=[0,2].
            First, the matrix should be reshped to (2,2,2,2,2,2)
            then we call  np.einsum('ijiiqi->jq', reshaped_dm)
        """
        return False

    @property
    def name(self):
        return self.__class__.__name__

    def __repr__(self):
        return f"Density Matrix"

    def hadamard(
        self,
        wires: Union[List[int], int],
        inverse: bool = False,
        comp_method: str = "bmm",
    ):
        dfunc.hadamard(self, wires=wires, inverse=inverse, comp_method=comp_method)

    def shadamard(
        self,
        wires: Union[List[int], int],
        inverse: bool = False,
        comp_method: str = "bmm",
    ):
        dfunc.shadamard(self, wires=wires, inverse=inverse, comp_method=comp_method)

    def paulix(
        self,
        wires: Union[List[int], int],
        inverse: bool = False,
        comp_method: str = "bmm",
    ):
        dfunc.paulix(self, wires=wires, inverse=inverse, comp_method=comp_method)

    def pauliy(
        self,
        wires: Union[List[int], int],
        inverse: bool = False,
        comp_method: str = "bmm",
    ):
        dfunc.pauliy(self, wires=wires, inverse=inverse, comp_method=comp_method)

    def pauliz(
        self,
        wires: Union[List[int], int],
        inverse: bool = False,
        comp_method: str = "bmm",
    ):
        dfunc.pauliz(self, wires=wires, inverse=inverse, comp_method=comp_method)

    def i(
        self,
        wires: Union[List[int], int],
        inverse: bool = False,
        comp_method: str = "bmm",
    ):
        dfunc.i(self, wires=wires, inverse=inverse, comp_method=comp_method)

    def s(
        self,
        wires: Union[List[int], int],
        inverse: bool = False,
        comp_method: str = "bmm",
    ):
        dfunc.s(self, wires=wires, inverse=inverse, comp_method=comp_method)

    def t(
        self,
        wires: Union[List[int], int],
        inverse: bool = False,
        comp_method: str = "bmm",
    ):
        dfunc.t(self, wires=wires, inverse=inverse, comp_method=comp_method)

    def sx(
        self,
        wires: Union[List[int], int],
        inverse: bool = False,
        comp_method: str = "bmm",
    ):
        dfunc.sx(self, wires=wires, inverse=inverse, comp_method=comp_method)

    def cnot(
        self,
        wires: Union[List[int], int],
        inverse: bool = False,
        comp_method: str = "bmm",
    ):
        dfunc.cnot(self, wires=wires, inverse=inverse, comp_method=comp_method)

    def cz(
        self,
        wires: Union[List[int], int],
        inverse: bool = False,
        comp_method: str = "bmm",
    ):
        dfunc.cz(self, wires=wires, inverse=inverse, comp_method=comp_method)

    def cy(
        self,
        wires: Union[List[int], int],
        inverse: bool = False,
        comp_method: str = "bmm",
    ):
        dfunc.cy(self, wires=wires, inverse=inverse, comp_method=comp_method)

    def swap(
        self,
        wires: Union[List[int], int],
        inverse: bool = False,
        comp_method: str = "bmm",
    ):
        dfunc.swap(self, wires=wires, inverse=inverse, comp_method=comp_method)

    def sswap(
        self,
        wires: Union[List[int], int],
        inverse: bool = False,
        comp_method: str = "bmm",
    ):
        dfunc.sswap(self, wires=wires, inverse=inverse, comp_method=comp_method)

    def cswap(
        self,
        wires: Union[List[int], int],
        inverse: bool = False,
        comp_method: str = "bmm",
    ):
        dfunc.cswap(self, wires=wires, inverse=inverse, comp_method=comp_method)

    def toffoli(
        self,
        wires: Union[List[int], int],
        inverse: bool = False,
        comp_method: str = "bmm",
    ):
        dfunc.toffoli(self, wires=wires, inverse=inverse, comp_method=comp_method)

    def multicnot(
        self,
        wires: Union[List[int], int],
        inverse: bool = False,
        comp_method: str = "bmm",
    ):

        dfunc.multicnot(self, wires=wires, inverse=inverse, comp_method=comp_method)

    def multixcnot(
        self,
        wires: Union[List[int], int],
        inverse: bool = False,
        comp_method: str = "bmm",
    ):

        dfunc.multixcnot(self, wires=wires, inverse=inverse, comp_method=comp_method)

    def rx(
        self,
        wires: Union[List[int], int],
        params: Union[torch.Tensor, np.ndarray, List[float], List[int], int, float],
        inverse: bool = False,
        comp_method: str = "bmm",
    ):

        if isinstance(params, Iterable):
            params = torch.Tensor(params)
        else:
            params = torch.Tensor([params])

        dfunc.rx(
            self, wires=wires, params=params, inverse=inverse, comp_method=comp_method
        )

    def ry(
        self,
        wires: Union[List[int], int],
        params: torch.Tensor,
        inverse: bool = False,
        comp_method: str = "bmm",
    ):

        if isinstance(params, Iterable):
            params = torch.Tensor(params)
        else:
            params = torch.Tensor([params])

        dfunc.ry(
            self, wires=wires, params=params, inverse=inverse, comp_method=comp_method
        )

    def rz(
        self,
        wires: Union[List[int], int],
        params: torch.Tensor,
        inverse: bool = False,
        comp_method: str = "bmm",
    ):

        if isinstance(params, Iterable):
            params = torch.Tensor(params)
        else:
            params = torch.Tensor([params])

        dfunc.rz(
            self, wires=wires, params=params, inverse=inverse, comp_method=comp_method
        )

    def rxx(
        self,
        wires: Union[List[int], int],
        params: Union[torch.Tensor, np.ndarray, List[float], List[int], int, float],
        inverse: bool = False,
        comp_method: str = "bmm",
    ):

        if isinstance(params, Iterable):
            params = torch.Tensor(params)
        else:
            params = torch.Tensor([params])

        dfunc.rxx(
            self, wires=wires, params=params, inverse=inverse, comp_method=comp_method
        )

    def ryy(
        self,
        wires: Union[List[int], int],
        params: Union[torch.Tensor, np.ndarray, List[float], List[int], int, float],
        inverse: bool = False,
        comp_method: str = "bmm",
    ):

        if isinstance(params, Iterable):
            params = torch.Tensor(params)
        else:
            params = torch.Tensor([params])

        dfunc.ryy(
            self, wires=wires, params=params, inverse=inverse, comp_method=comp_method
        )

    def rzz(
        self,
        wires: Union[List[int], int],
        params: Union[torch.Tensor, np.ndarray, List[float], List[int], int, float],
        inverse: bool = False,
        comp_method: str = "bmm",
    ):

        if isinstance(params, Iterable):
            params = torch.Tensor(params)
        else:
            params = torch.Tensor([params])

        dfunc.rzz(
            self, wires=wires, params=params, inverse=inverse, comp_method=comp_method
        )

    def rzx(
        self,
        wires: Union[List[int], int],
        params: Union[torch.Tensor, np.ndarray, List[float], List[int], int, float],
        inverse: bool = False,
        comp_method: str = "bmm",
    ):

        if isinstance(params, Iterable):
            params = torch.Tensor(params)
        else:
            params = torch.Tensor([params])

        dfunc.rzx(
            self, wires=wires, params=params, inverse=inverse, comp_method=comp_method
        )

    def phaseshift(
        self,
        wires: Union[List[int], int],
        params: Union[torch.Tensor, np.ndarray, List[float], List[int], int, float],
        inverse: bool = False,
        comp_method: str = "bmm",
    ):

        if isinstance(params, Iterable):
            params = torch.Tensor(params)
        else:
            params = torch.Tensor([params])

        dfunc.phaseshift(
            self, wires=wires, params=params, inverse=inverse, comp_method=comp_method
        )

    def rot(
        self,
        wires: Union[List[int], int],
        params: Union[torch.Tensor, np.ndarray, List[float], List[int], int, float],
        inverse: bool = False,
        comp_method: str = "bmm",
    ):

        if isinstance(params, Iterable):
            params = torch.Tensor(params)
        else:
            params = torch.Tensor([params])

        if params.dim() == 1:
            params = params.unsqueeze(0)

        dfunc.rot(
            self, wires=wires, params=params, inverse=inverse, comp_method=comp_method
        )

    def multirz(
        self,
        wires: Union[List[int], int],
        params: Union[torch.Tensor, np.ndarray, List[float], List[int], int, float],
        inverse: bool = False,
        comp_method: str = "bmm",
    ):

        if isinstance(params, Iterable):
            params = torch.Tensor(params)
        else:
            params = torch.Tensor([params])

        dfunc.multirz(
            self, wires=wires, params=params, inverse=inverse, comp_method=comp_method
        )

    def crx(
        self,
        wires: Union[List[int], int],
        params: Union[torch.Tensor, np.ndarray, List[float], List[int], int, float],
        inverse: bool = False,
        comp_method: str = "bmm",
    ):

        if isinstance(params, Iterable):
            params = torch.Tensor(params)
        else:
            params = torch.Tensor([params])

        dfunc.crx(
            self, wires=wires, params=params, inverse=inverse, comp_method=comp_method
        )

    def cry(
        self,
        wires: Union[List[int], int],
        params: Union[torch.Tensor, np.ndarray, List[float], List[int], int, float],
        inverse: bool = False,
        comp_method: str = "bmm",
    ):

        if isinstance(params, Iterable):
            params = torch.Tensor(params)
        else:
            params = torch.Tensor([params])

        dfunc.cry(
            self, wires=wires, params=params, inverse=inverse, comp_method=comp_method
        )

    def crz(
        self,
        wires: Union[List[int], int],
        params: Union[torch.Tensor, np.ndarray, List[float], List[int], int, float],
        inverse: bool = False,
        comp_method: str = "bmm",
    ):

        if isinstance(params, Iterable):
            params = torch.Tensor(params)
        else:
            params = torch.Tensor([params])

        dfunc.crz(
            self, wires=wires, params=params, inverse=inverse, comp_method=comp_method
        )

    def crot(
        self,
        wires: Union[List[int], int],
        params: Union[torch.Tensor, np.ndarray, List[float], List[int], int, float],
        inverse: bool = False,
        comp_method: str = "bmm",
    ):

        if isinstance(params, Iterable):
            params = torch.Tensor(params)
        else:
            params = torch.Tensor([params])

        if params.dim() == 1:
            params = params.unsqueeze(0)

        dfunc.crot(
            self, wires=wires, params=params, inverse=inverse, comp_method=comp_method
        )

    def u1(
        self,
        wires: Union[List[int], int],
        params: Union[torch.Tensor, np.ndarray, List[float], List[int], int, float],
        inverse: bool = False,
        comp_method: str = "bmm",
    ):

        if isinstance(params, Iterable):
            params = torch.Tensor(params)
        else:
            params = torch.Tensor([params])

        dfunc.u1(
            self, wires=wires, params=params, inverse=inverse, comp_method=comp_method
        )

    def u2(
        self,
        wires: Union[List[int], int],
        params: Union[torch.Tensor, np.ndarray, List[float], List[int], int, float],
        inverse: bool = False,
        comp_method: str = "bmm",
    ):

        if isinstance(params, Iterable):
            params = torch.Tensor(params)
        else:
            params = torch.Tensor([params])

        if params.dim() == 1:
            params = params.unsqueeze(0)

        dfunc.u2(
            self, wires=wires, params=params, inverse=inverse, comp_method=comp_method
        )

    def u3(
        self,
        wires: Union[List[int], int],
        params: Union[torch.Tensor, np.ndarray, List[float], List[int], int, float],
        inverse: bool = False,
        comp_method: str = "bmm",
    ):

        if isinstance(params, Iterable):
            params = torch.Tensor(params)
        else:
            params = torch.Tensor([params])

        if params.dim() == 1:
            params = params.unsqueeze(0)

        dfunc.u3(
            self, wires=wires, params=params, inverse=inverse, comp_method=comp_method
        )

    def cu1(
        self,
        wires: Union[List[int], int],
        params: Union[torch.Tensor, np.ndarray, List[float], List[int], int, float],
        inverse: bool = False,
        comp_method: str = "bmm",
    ):

        if isinstance(params, Iterable):
            params = torch.Tensor(params)
        else:
            params = torch.Tensor([params])

        if params.dim() == 1:
            params = params.unsqueeze(0)

        dfunc.cu1(
            self, wires=wires, params=params, inverse=inverse, comp_method=comp_method
        )

    def cu2(
        self,
        wires: Union[List[int], int],
        params: Union[torch.Tensor, np.ndarray, List[float], List[int], int, float],
        inverse: bool = False,
        comp_method: str = "bmm",
    ):

        if isinstance(params, Iterable):
            params = torch.Tensor(params)
        else:
            params = torch.Tensor([params])

        if params.dim() == 1:
            params = params.unsqueeze(0)

        dfunc.cu2(
            self, wires=wires, params=params, inverse=inverse, comp_method=comp_method
        )

    def cu3(
        self,
        wires: Union[List[int], int],
        params: Union[torch.Tensor, np.ndarray, List[float], List[int], int, float],
        inverse: bool = False,
        comp_method: str = "bmm",
    ):

        if isinstance(params, Iterable):
            params = torch.Tensor(params)
        else:
            params = torch.Tensor([params])

        if params.dim() == 1:
            params = params.unsqueeze(0)

        dfunc.cu3(
            self, wires=wires, params=params, inverse=inverse, comp_method=comp_method
        )

    def qubitunitary(
        self,
        wires: Union[List[int], int],
        params: Union[torch.Tensor, np.ndarray, List[float], List[int], int, float],
        inverse: bool = False,
        comp_method: str = "bmm",
    ):

        if isinstance(params, Iterable):
            params = torch.tensor(params, dtype=C_DTYPE)
        else:
            params = torch.tensor([params], dtype=C_DTYPE)

        dfunc.qubitunitary(
            self, wires=wires, params=params, inverse=inverse, comp_method=comp_method
        )

    def qubitunitaryfast(
        self,
        wires: Union[List[int], int],
        params: Union[torch.Tensor, np.ndarray, List[float], List[int], int, float],
        inverse: bool = False,
        comp_method: str = "bmm",
    ):

        if isinstance(params, Iterable):
            params = torch.tensor(params, dtype=C_DTYPE)
        else:
            params = torch.tensor([params], dtype=C_DTYPE)

        dfunc.qubitunitaryfast(
            self, wires=wires, params=params, inverse=inverse, comp_method=comp_method
        )

    def qubitunitarystrict(
        self,
        wires: Union[List[int], int],
        params: Union[torch.Tensor, np.ndarray, List[float], List[int], int, float],
        inverse: bool = False,
        comp_method: str = "bmm",
    ):

        if isinstance(params, Iterable):
            params = torch.tensor(params, dtype=C_DTYPE)
        else:
            params = torch.tensor([params], dtype=C_DTYPE)

        dfunc.qubitunitarystrict(
            self, wires=wires, params=params, inverse=inverse, comp_method=comp_method
        )

    def single_excitation(
        self,
        wires: Union[List[int], int],
        params: Union[torch.Tensor, np.ndarray, List[float], List[int], int, float],
        inverse: bool = False,
        comp_method: str = "bmm",
    ):

        if isinstance(params, Iterable):
            params = torch.Tensor(params)
        else:
            params = torch.Tensor([params])

        dfunc.single_excitation(
            self, wires=wires, params=params, inverse=inverse, comp_method=comp_method
        )

    h = hadamard
    sh = shadamard
    x = paulix
    y = pauliy
    z = pauliz
    xx = rxx
    yy = ryy
    zz = rzz
    zx = rzx
    cx = cnot
    ccnot = toffoli
    ccx = toffoli
    u = u3
    cu = cu3
    p = phaseshift
    cp = cu1
    cr = cu1
    cphase = cu



import copy
import time
from typing import List, Any, Dict


import torch
import torch.nn.functional as F
import tqdm
from torch.utils.data import DataLoader

from torchpack.callbacks.callback import Callback, Callbacks
from torchpack.utils import humanize
from torchpack.utils.logging import logger
from torchpack.utils.typing import Trainer
from torchpack import distributed as dist
from torchquantum.super_utils import get_named_sample_arch
from torchquantum.utils import legalize_unitary


__all__ = [
    "LegalInferenceRunner",
    "SubnetInferenceRunner",
    "NLLError",
    "TrainerRestore",
    "MinError",
    "AddNoiseInferenceRunner",
    "GradRestore",
]


class LegalInferenceRunner(Callback):
    """
    A callback that runs inference with a list of :class:`Callback`.
    """

    def __init__(self, dataflow: DataLoader, *, callbacks: List[Callback]) -> None:
        self.dataflow = dataflow
        self.callbacks = Callbacks(callbacks)

    def _set_trainer(self, trainer: Trainer) -> None:
        self.callbacks.set_trainer(trainer)

    def _trigger_epoch(self) -> None:
        self._trigger()

    def _trigger(self) -> None:
        start_time = time.perf_counter()
        self.callbacks.before_epoch()

        with torch.no_grad():
            self.trainer.legalized_model = copy.deepcopy(self.trainer.model)
            legalize_unitary(self.trainer.legalized_model)
            for feed_dict in tqdm.tqdm(self.dataflow, ncols=0):
                self.callbacks.before_step(feed_dict)
                output_dict = self.trainer.run_step(feed_dict, legalize=True)
                self.callbacks.after_step(output_dict)

        self.callbacks.after_epoch()
        logger.info(
            "Inference finished in {}.".format(
                humanize.naturaldelta(time.perf_counter() - start_time)
            )
        )


class SubnetInferenceRunner(Callback):
    """
    A callback that runs inference with a list of :class:`Callback`.
    sample a subnet and run during supernet training
    """

    def __init__(
        self, dataflow: DataLoader, *, callbacks: List[Callback], subnet: str
    ) -> None:
        self.dataflow = dataflow
        self.callbacks = Callbacks(callbacks)
        self.subnet = subnet

    def _set_trainer(self, trainer: Trainer) -> None:
        self.callbacks.set_trainer(trainer)

    def _trigger_epoch(self) -> None:
        self._trigger()

    def _trigger(self) -> None:
        start_time = time.perf_counter()
        self.callbacks.before_epoch()

        with torch.no_grad():
            sample_arch = get_named_sample_arch(
                self.trainer.model.arch_space, self.subnet
            )
            self.trainer.model.set_sample_arch(sample_arch)
            for feed_dict in tqdm.tqdm(self.dataflow, ncols=0):
                self.callbacks.before_step(feed_dict)
                output_dict = self.trainer.run_step(feed_dict)
                self.callbacks.after_step(output_dict)

        self.callbacks.after_epoch()
        logger.info(
            "Inference finished in {}.".format(
                humanize.naturaldelta(time.perf_counter() - start_time)
            )
        )


class AddNoiseInferenceRunner(Callback):
    """
    A callback that runs inference with a list of :class:`Callback`.
    sample noise and add to model during training
    """

    def __init__(
        self,
        dataflow: DataLoader,
        *,
        callbacks: List[Callback],
        noise_total_prob: float
    ) -> None:
        self.dataflow = dataflow
        self.callbacks = Callbacks(callbacks)
        self.noise_total_prob = noise_total_prob

    def _set_trainer(self, trainer: Trainer) -> None:
        self.callbacks.set_trainer(trainer)

    def _trigger_epoch(self) -> None:
        self._trigger()

    def _trigger(self) -> None:
        start_time = time.perf_counter()
        self.callbacks.before_epoch()

        with torch.no_grad():
            orig_is_add_noise = self.trainer.model.nodes[0].noise_model_tq.is_add_noise
            orig_noise_total_prob = self.trainer.model.nodes[
                0
            ].noise_model_tq.noise_total_prob
            orig_mode = self.trainer.model.nodes[0].noise_model_tq.mode

            for node in self.trainer.model.nodes:
                node.noise_model_tq.noise_total_prob = self.noise_total_prob
                node.noise_model_tq.is_add_noise = True
                node.noise_model_tq.mode = "train"

            for feed_dict in tqdm.tqdm(self.dataflow, ncols=0):
                self.callbacks.before_step(feed_dict)
                output_dict = self.trainer.run_step(feed_dict)
                self.callbacks.after_step(output_dict)

            for node in self.trainer.model.nodes:
                node.noise_model_tq.is_add_noise = orig_is_add_noise
                node.noise_model_tq.noise_total_prob = orig_noise_total_prob
                node.noise_model_tq.mode = orig_mode

        self.callbacks.after_epoch()
        logger.info(
            "Inference finished in {}.".format(
                humanize.naturaldelta(time.perf_counter() - start_time)
            )
        )


class NLLError(Callback):
    def __init__(
        self,
        *,
        output_tensor: str = "outputs",
        target_tensor: str = "targets",
        name: str = "error"
    ) -> None:
        self.output_tensor = output_tensor
        self.target_tensor = target_tensor
        self.name = name

    def _before_epoch(self):
        self.size = 0
        self.errors = 0

    def _after_step(self, output_dict: Dict[str, Any]) -> None:
        outputs = output_dict[self.output_tensor]
        targets = output_dict[self.target_tensor]

        error = F.nll_loss(outputs, targets)

        self.size += targets.size(0)
        self.errors += error.item() * targets.size(0)

    def _after_epoch(self) -> None:
        self.size = dist.allreduce(self.size, reduction="sum")
        self.errors = dist.allreduce(self.errors, reduction="sum")
        self.trainer.summary.add_scalar(self.name, self.errors / self.size)


class MinError(Callback):
    def __init__(
        self,
        *,
        output_tensor: str = "outputs",
        target_tensor: str = "targets",
        name: str = "error"
    ) -> None:
        self.output_tensor = output_tensor
        self.target_tensor = target_tensor
        self.name = name

    def _before_epoch(self):
        self.size = 0
        self.errors = 0

    def _after_step(self, output_dict: Dict[str, Any]) -> None:
        outputs = output_dict[self.output_tensor]
        targets = output_dict[self.target_tensor]

        error = outputs.sum()

        self.size += outputs.size(0)
        self.errors += error.item()

    def _after_epoch(self) -> None:
        self.size = dist.allreduce(self.size, reduction="sum")
        self.errors = dist.allreduce(self.errors, reduction="sum")
        self.trainer.summary.add_scalar(self.name, self.errors / self.size)


class TrainerRestore(Callback):
    def __init__(self, state) -> None:
        self.state = state

    def _before_train(self) -> None:
        self.trainer.load_state_dict(self.state)


class GradRestore(Callback):
    """
    A callback that restore the all the gradients among all the steps.
    """

    def __init__(self) -> None:
        self.trainer = None
        pass

    def _set_trainer(self, trainer: Trainer) -> None:
        self.trainer = trainer

    def _trigger_step(self) -> None:
        self._trigger()

    def _trigger(self) -> None:
        for node in self.trainer.model.nodes:
            for i, param in enumerate(node.q_layer.parameters()):
                self.trainer.summary.add_scalar(
                    "grad/grad_" + str(i), float(param.grad)
                )
                self.trainer.summary.add_scalar("param/param_" + str(i), float(param))
                # self.trainer.summary.writers[1].add_histogram('histogram/grad', float(param.grad))



import argparse
import os
import sys
import pdb
import numpy as np
import torch
import torch.backends.cudnn
import torch.cuda
import torch.nn
import torch.utils.data
import torchquantum as tq

# from torchpack import distributed as dist
from torchpack.environ import set_run_dir
from torchpack.utils.config import configs
from torchpack.utils.logging import logger
from examples.vqe import builder
from torchquantum.vqe_utils import parse_hamiltonian_file
from examples.vqe.q_models import QVQEModel0


def main() -> None:
    # dist.init()
    torch.backends.cudnn.benchmark = True
    # torch.cuda.set_device(dist.local_rank())

    parser = argparse.ArgumentParser()
    parser.add_argument("config", metavar="FILE", help="config file")
    parser.add_argument("--ckpt-dir", metavar="DIR", help="run directory")
    parser.add_argument("--pdb", action="store_true", help="pdb")
    parser.add_argument("--gpu", type=str, help="gpu ids", default=None)
    parser.add_argument(
        "--print-configs", action="store_true", help="print ALL configs"
    )
    args, opts = parser.parse_known_args()

    configs.load(args.config, recursive=True)
    configs.update(opts)

    if configs.debug.pdb or args.pdb:
        pdb.set_trace()

    if args.gpu is not None:
        os.environ["CUDA_VISIBLE_DEVICES"] = args.gpu

    if configs.debug.set_seed:
        torch.manual_seed(configs.debug.seed)
        np.random.seed(configs.debug.seed)

    if configs.run.device == "gpu":
        if torch.cuda.is_available():
            device = torch.device("cuda")
        else:
            configs.run.device = "cpu"
            device = torch.device("cpu")
    elif configs.run.device == "cpu":
        device = torch.device("cpu")
    else:
        raise ValueError(configs.run.device)

    if isinstance(configs.optimizer.lr, str):
        configs.optimizer.lr = eval(configs.optimizer.lr)

    # set the run dir according to config file's name
    args.run_dir = "runs/" + args.config.replace("/", ".").replace(
        "examples.", ""
    ).replace(".yml", "").replace("configs.", "")
    set_run_dir(args.run_dir)

    logger.info(" ".join([sys.executable] + sys.argv))

    configs.model.hamil_filename = "examples/vqe/h2.txt"

    print_conf = configs

    logger.info(f'Training started: "{args.run_dir}".' + "\n" + f"{print_conf}")

    from torchquantum.datasets import VQE

    dataset = VQE(steps_per_epoch=configs.run.steps_per_epoch)
    dataflow = dict()

    for split in dataset:
        if split == "train":
            sampler = torch.utils.data.RandomSampler(dataset[split])
        else:
            # for valid and test, use SequentialSampler to make the train.py
            # and eval.py results consistent
            sampler = torch.utils.data.SequentialSampler(dataset[split])

        dataflow[split] = torch.utils.data.DataLoader(
            dataset[split],
            batch_size=configs.run.bsz,
            sampler=sampler,
            num_workers=configs.run.workers_per_gpu,
            pin_memory=True,
        )

    model = QVQEModel0(
        arch=configs.model.arch,
        hamil_info=parse_hamiltonian_file(configs.model.hamil_filename),
    )
    model.to(device)

    total_params = sum(p.numel() for p in model.parameters())
    logger.info(f"Model Size: {total_params}")

    # make criterion
    from examples.vqe.criterions import minimize

    criterion = minimize

    optimizer = builder.make_optimizer(model)
    scheduler = builder.make_scheduler(optimizer)
    trainer = builder.make_trainer(
        model=model, criterion=criterion, optimizer=optimizer, scheduler=scheduler
    )

    # trainer state_dict will be loaded in a callback
    callbacks = builder.make_callbacks(dataflow)
    trainer.set_use_qiskit(configs)
    if configs.qiskit.use_qiskit_train or configs.qiskit.use_qiskit_valid:
        from torchquantum.plugins import QiskitProcessor

        processor = QiskitProcessor(
            use_real_qc=configs.qiskit.use_real_qc,
            n_shots=configs.qiskit.n_shots,
            backend_name=configs.qiskit.backend_name,
        )
        model.set_qiskit_processor(processor)

    trainer.train_with_defaults(
        dataflow["train"], num_epochs=configs.run.n_epochs, callbacks=callbacks
    )


if __name__ == "__main__":
    main()



import qiskit
import os
import shutil

if __name__ == "__main__":
    path = os.path.abspath(qiskit.__file__)
    print(path)
    # path for aer provider
    path_provider = path.replace("__init__.py", "providers/aer/backends/aerbackend.py")
    print(path_provider)
    fixed_file = "aerbackend_fixed.py"

    with open(path_provider, "r") as fid:
        for line in fid.readlines():
            if "FIXED" in line:
                print("The qiskit parameterization bug is already fixed!")
                exit(0)
            else:
                print(
                    f"Fixing the qiskit parameterization bug by replacing "
                    f"the {path_provider} file with {fixed_file}!"
                )
                break

    shutil.copyfile(
        path_provider, path_provider.replace("aerbackend", "aerbackend.orig")
    )
    shutil.copyfile(fixed_file, path_provider)
    print("Finished!")



import math
import torch
import numpy as np

from string import ascii_lowercase

torch_numpy_dtype_dict = {
    torch.complex64: np.complex64,
    torch.complex128: np.complex128,
    torch.float32: np.float32,
    torch.float64: np.float64,
    torch.int32: np.int32,
    torch.int64: np.int64,
    torch.bool: np.bool_,
}

C_DTYPE = torch.complex64
F_DTYPE = torch.float32

ABC = ascii_lowercase
ABC_ARRAY = np.array(list(ABC))

INV_SQRT2 = 1 / math.sqrt(2)

C_DTYPE_NUMPY = torch_numpy_dtype_dict[C_DTYPE]
F_DTYPE_NUMPY = torch_numpy_dtype_dict[F_DTYPE]



__version__ = "0.1.7"
__author__ = "Hanrui Wang, Jiannan Cao, Jessica Ding, Jiai Gu, Song Han, Zirui Li, Zhiding Liang, Pengyu Liu, Mohammadreza Tavasoli"

from .macro import *
from .devices import *
from .module import *
from .operators import *
from .measurement import *
from .functional import *
from .graph import *
from .layers import *
from .encoding import *
from .super_layers import *
from .utils import *
from .noise_model import *
from .node import *
from .pulse import *

# from .pulse_utils import *

# here we check whether the Qiskit parameterization bug is fixed, if not, a
# warning message will be printed
import qiskit
import os

path = os.path.abspath(qiskit.__file__)
# print(path)
# path for aer provider
path_provider = path.replace("__init__.py", "providers/aer/backends/aerbackend.py")
# print(path_provider)

# with open(path_provider, 'r') as fid:
#     for line in fid.readlines():
#         if 'FIXED' in line:
#             # print('The qiskit parameterization bug is already fixed!')
#             break
#         else:
#             print(f'\n\n WARNING: The qiskit parameterization bug is not '
#                   f'fixed!\n\n'
#                   f'run python fix_qiskit_parameterization.py to fix it!'
#                   )
#             break
